@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*  GMime
- *  Copyright (C) 2000-2009 Jeffrey Stedfast
+ *  Copyright (C) 2000-2007 Jeffrey Stedfast
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -175,7 +175,7 @@ test_stream_buffer_gets (const char *filename)
 		testsuite_check_failed ("GMimeStreamBuffer::block gets() failed: %s",
 					ex->message);
 	} finally {
-		g_object_unref (buffered);
+		g_mime_stream_unref (buffered);
 	}
 	
 	testsuite_check ("GMimeStreamBuffer::cache gets");
@@ -188,10 +188,10 @@ test_stream_buffer_gets (const char *filename)
 		testsuite_check_failed ("GMimeStreamBuffer::block gets() failed: %s",
 					ex->message);
 	} finally {
-		g_object_unref (buffered);
+		g_mime_stream_unref (buffered);
 	}
 	
-	g_object_unref (stream);
+	g_mime_stream_unref (stream);
 }
 
 
@@ -229,8 +229,8 @@ test_stream_mem (const char *filename)
 					ex->message);
 	} finally;
 	
-	g_object_unref (fstream);
-	g_object_unref (stream);
+	g_mime_stream_unref (fstream);
+	g_mime_stream_unref (stream);
 	
 	testsuite_end ();
 }
@@ -484,8 +484,9 @@ static void
 gen_random_stream (GMimeStream *stream)
 {
 	size_t nwritten, buflen, total = 0, size, i;
-	char buf[4096];
+	char buf[4096], *p;
 	ssize_t n;
+	int fd;
 	
 	/* read between 4k and 14k bytes */
 	size = 4096 + (size_t) (10240.0 * (rand () / (RAND_MAX + 1.0)));
@@ -516,13 +517,13 @@ gen_random_stream (GMimeStream *stream)
 }
 
 static int
-gen_test_data (const char *datadir, char **stream_name)
+gen_test_data (const char *datadir)
 {
 	GMimeStream *istream, *ostream, *stream;
 	char input[256], output[256], *name, *p;
+	size_t left, len;
 	off_t start, end;
 	struct stat st;
-	size_t len;
 	int fd, i;
 	
 	srand (time (NULL));
@@ -544,8 +545,6 @@ gen_test_data (const char *datadir, char **stream_name)
 	if ((fd = mkstemp (input)) == -1)
 		return -1;
 	
-	*stream_name = g_strdup (name);
-	
 	*p++ = G_DIR_SEPARATOR;
 	p = g_stpcpy (p, name);
 	*p++ = ':';
@@ -590,10 +589,10 @@ int main (int argc, char **argv)
 {
 	const char *datadir = "data/streams";
 	gboolean gen_data = TRUE;
-	char *stream_name = NULL;
 	struct dirent *dent;
 	char path[256], *p;
 	DIR *dir, *outdir;
+	struct stat st;
 	int i;
 	
 	g_mime_init (0);
@@ -614,7 +613,7 @@ int main (int argc, char **argv)
 	strcpy (p, "output");
 	
 	if (!(outdir = opendir (path))) {
-		if (gen_test_data (datadir, &stream_name) == -1 ||
+		if (gen_test_data (datadir) == -1 ||
 		    !(outdir = opendir (path)))
 			goto exit;
 		
@@ -624,7 +623,7 @@ int main (int argc, char **argv)
 	p = g_stpcpy (p, "input");
 	
 	if (!(dir = opendir (path))) {
-		if (!gen_data || gen_test_data (datadir, &stream_name) == -1 ||
+		if (!gen_data || gen_test_data (datadir) == -1 ||
 		    !(dir = opendir (path))) {
 			closedir (outdir);
 			goto exit;
@@ -644,7 +643,7 @@ int main (int argc, char **argv)
 		
 		rewinddir (dir);
 		
-		if (gen_data && gen_test_data (datadir, &stream_name) == -1)
+		if (gen_data && gen_test_data (datadir) == -1)
 			goto exit;
 	}
 	
@@ -661,27 +660,6 @@ int main (int argc, char **argv)
 		test_stream_buffer_gets (path);
 	}
 	
-	if (gen_data && stream_name && testsuite_total_errors () == 0) {
-		/* since all tests were successful, unlink the generated test data */
-		strcpy (p, stream_name);
-		unlink (path);
-		
-		p = g_stpcpy (path, datadir);
-		*p++ = G_DIR_SEPARATOR;
-		p = g_stpcpy (p, "output");
-		*p++ = G_DIR_SEPARATOR;
-		
-		rewinddir (outdir);
-		while ((dent = readdir (outdir))) {
-			if (!strncmp (dent->d_name, stream_name, strlen (stream_name))) {
-				strcpy (p, dent->d_name);
-				unlink (path);
-			}
-		}
-		
-		g_free (stream_name);
-	}
-	
 	closedir (outdir);
 	closedir (dir);
 	
